;;; -*- Mode: Emacs-Lisp -*-

;;;; Superior Lisp Interaction Mode for Emacs
;;;; Statistical Profiling Interface

;;; Copyright (C) 2005, 2006 Juho Snellman
;;; Copyright (C) 2006, 2007 Taylor R Campbell
;;;
;;; Permission is hereby granted, free of charge, to any person
;;; obtaining a copy of this software and associated documentation
;;; files (the "Software"), to deal in the Software without
;;; restriction, including without limitation the rights to use, copy,
;;; modify, merge, publish, distribute, sublicense, and/or sell copies
;;; of the Software, and to permit persons to whom the Software is
;;; furnished to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be
;;; included in all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
;;; BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
;;; ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
;;; CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
;;; SOFTWARE.

;;; This page must come first because it defines some functions needed
;;; by the custom stuff on the next page.

(defvar slime-sprofile-browser-column-descriptors
  '((0 id               "Id"
                        "Identifier known to Lisp"
                        slime-sprofile-browser-insert-integer
                        <)
    (1 name             "Name"
                        "Name of code block sampled"
                        slime-sprofile-browser-insert-name
                        string<)
    (2 self-time        "Self"
                        "Percent direct samples"
                        slime-sprofile-browser-insert-percent
                        >)              ;Order higher values first.
    (3 cumulative-time  "Cumulative"
                        "Percent indirect samples on stack"
                        slime-sprofile-browser-insert-percent
                        >)
    (4 total-time       "Total"
                        "I don't remember!"
                        slime-sprofile-browser-insert-percent
                        >))
  "List of descriptors for sprofile browser columns.
A descriptor is a list
  (INDEX NAME HEADER DESCRIPTION FORMATTER COMPARATOR),
where
- INDEX is the index of the field in a profile node from Lisp,
- NAME is a symbol,
- DESCRIPTION is a short phrase describing the field,
- FORMATTER is a function that takes a column's value and a width and
  inserts the column into the current buffer, and
- COMPARATOR is a binary ordering relation on the columns' values.")

(defun slime-sprofile-browser-column-index (column) (nth 0 column))
(defun slime-sprofile-browser-column-name (column) (nth 1 column))
(defun slime-sprofile-browser-column-header (column) (nth 2 column))
(defun slime-sprofile-browser-column-description (column)
  (nth 3 column))
(defun slime-sprofile-browser-column-formatter (column) (nth 4 column))
(defun slime-sprofile-browser-column-comparator (column)
  (nth 5 column))

(defun slime-sprofile-node-id (node) (car node))

(defun slime-sprofile-browser-lookup-column (name)
  (catch 'return
    (let ((columns slime-sprofile-browser-column-descriptors))
      (while (not (null columns))
        (let ((column (car columns)))
          (if (eq name (slime-sprofile-browser-column-name column))
              (throw 'return column)
            (setq columns (cdr columns))))))
    (error "No such sprofile browser column: %S" name)))

(defvar slime-sprofile-browser-column-choices
  (mapcar
   (lambda (column)
     `(const :tag ,(slime-sprofile-browser-column-description column)
             ,(slime-sprofile-browser-column-name column)))
   slime-sprofile-browser-column-descriptors))

;;;; Customization

(defgroup slime-sprofile-browser nil
  "SLIME Statistical Profile (sprofile) Browser"
  :prefix "slime-sprofile-browser-"
  :group 'slime)

;;; The only reason that this is customizable is that I couldn't live
;;; with myself having publicly released code that generated the
;;; columns with magic constants, and so I *had* to generalize it or
;;; something nasty might have happened.

(defcustom slime-sprofile-browser-column-layout
  '((name               left    variable)
    (self-time          right   6)
    (cumulative-time    right   6)
    (total-time         right   6))
  "*Layout of rows' columns in the sprofile browser.
This is a list of elements of the form (COLUMN-NAME POSITION WIDTH),
  where COLUMN-NAME is the symbol `id', `name', `self-time',
  `cumulative-time', or `total-time'; POSITION is the symbol `left',
  `right', or `centre', indicating justification for only the title of
  the entry; and WIDTH is either a non-negative integer or the symbol
  `variable'.
If WIDTH is an integer, then the column will occupy that many character
  cells.
If WIDTH is the symbol `variable', then the column will be expanded so
  that the whole display will fill the window."
  :type `(repeat
          (list (choice :tag "Column"
                        ,slime-sprofile-browser-column-choices)
                (choice :tag "Header justification"
                        (const :tag "Left-justified" left)
                        (const :tag "Right-justified" right)
                        (const :tag "Centred" centre))
                (choice :tag "Width"
                        (integer :tag "Number of character cells")
                        (const :tag "Variable width" variable)))))

(defcustom slime-sprofile-browser-order-column 'self-time
  "*Column by which to sort sprofile browser listings."
  :type `(choice ,@slime-sprofile-browser-column-choices))
(make-variable-buffer-local 'slime-sprofile-browser-order-column)

;;;; SLIME Commands

(defun slime-sprofile-start ()
  "Start sampling the Lisp image.."
  (interactive)
  (slime-eval '(swank:swank-sprofile-start)))

(defun slime-sprofile-stop ()
  "Stop sampling the Lisp image."
  (interactive)
  (slime-eval '(swank:swank-sprofile-stop)))

(defun slime-sprofile-browse ()
  "Browse the statistical profile after sampling.
Note that this does not stop the sampling.  Type \[slime-sprofile-stop]
  to stop sampling, or \[slime-sprofile-stop-and-browse] to stop
  sampling and browse browse."
  (interactive)
  (slime-eval-async '(swank:swank-sprofile-get-call-graph)
    (lambda (graph)
      (switch-to-buffer (slime-sprofile-browser-get-buffer))
      (setq slime-sprofile-browser-graph graph)
      (slime-sprofile-browser-mode))))

(defun slime-sprofile-stop-and-browse ()
  "Stop sampling and browse the statistical profile."
  (interactive)
  (slime-sprofile-stop)
  (slime-sprofile-browse))

(defslime-repl-shortcut slime-repl-shortcut-start-sprofile
    ("start-sprofile" "sample")
  (:handler 'slime-sprofile-start)
  (:one-liner "Start sampling for a statistical profile."))

(defslime-repl-shortcut slime-repl-shortcut-stop-sprofile
    ("stop-sprofile")
  (:handler 'slime-sprofile-stop)
  (:one-liner "Stop sampling for a statistical profile."))

(defslime-repl-shortcut slime-repl-shortcut-browse-sprofile
    ("browse-sprofile")
  (:handler 'slime-sprofile-browse)
  (:one-liner "Browse the statistical profile."))

(defslime-repl-shortcut slime-repl-shortcut-stop-and-browse-sprofile
    ("stop-and-browse-sprofile" "sbp")
  (:handler 'slime-sprofile-stop-and-browse)
  (:one-liner "Stop sampling and browse the statistical profile."))

;;;; Browser

(defvar slime-sprofile-browser-graph nil
  "Graph of statistical profiling results.
A list of entries of the form (INDEX NAME SELF CUMULATIVE TOTAL).
Set in sprofile browser buffers before entering the sprofile browser
  mode.")

(defvar slime-sprofile-browser-fill-width nil
  "Width by which to fill variable columns.
Cached according to the window width.")
(make-variable-buffer-local 'slime-sprofile-browser-fill-width)

(define-derived-mode slime-sprofile-browser-mode fundamental-mode
  "slprof"
  "Major mode for browsing statistical profiler data.

\\{slime-sprofile-browser-mode-map}"
  (setq buffer-read-only t)
  (slime-sprofile-browser-redraw))

(slime-define-keys slime-sprofile-browser-mode-map
  ("d" 'slime-sprofile-browser-disassemble)
  ((kbd "DEL") 'scroll-down)
  ("f" 'slime-sprofile-browser-follow)
  ("g" 'slime-sprofile-browser-redraw)  ;Convention (`G'et mail?)
  ("q" 'slime-temp-buffer-quit)
  ((kbd "RET") 'slime-sprofile-browser-toggle)
  ("s" 'slime-sprofile-browser-sort)
  ((kbd "SPC") 'scroll-up)
  ("v" 'slime-sprofile-browser-view-source))

(defun slime-sprofile-browser-get-buffer ()
  (get-buffer-create "*slime-sprofile*"))

(defun slime-sprofile-browser-check ()
  (if (not (eq major-mode 'slime-sprofile-browser-mode))
      (error "Not in SLIME sprofile browser")))

(defvar slime-sprofile-browser-order-index nil
  "Dynamic variable for sorting the sprofile graph.")

(defvar slime-sprofile-browser-order-comparator nil
  "Dynamic variable for sorting the sprofile graph.")

(defun slime-sprofile-browser-sort-graph (graph)
  (let ((name slime-sprofile-browser-order-column))
    (if (not name)
        graph
        (let ((column
               (slime-sprofile-browser-lookup-column name)))
          (let ((slime-sprofile-browser-order-index
                 (slime-sprofile-browser-column-index column))
                (slime-sprofile-browser-order-comparator
                 (slime-sprofile-browser-column-comparator column)))
            (sort graph 'slime-sprofile-browser-compare-nodes))))))

(defun slime-sprofile-browser-compare-nodes (node-a node-b)
  (funcall slime-sprofile-browser-order-comparator
           (nth slime-sprofile-browser-order-index node-a)
           (nth slime-sprofile-browser-order-index node-b)))

;;;;; Browser Drawing

(defun slime-sprofile-browser-calculate-fill ()
  "Calculate the width of the variable sprofile column, if any."
  (let ((fixed-width 0)        ;Number of cells used by other columns
        (variable nil))        ;The variable column, if we have seen it
    (dolist (layout slime-sprofile-browser-column-layout)
      (let ((name (car layout))
            (width (caddr layout)))
        (if (eq width 'variable)
            (if variable
                (error "Two sprofile columns are variable: %S and %S"
                       variable
                       name)
                (setq variable name))
            ;; Add one for a space.
            (setq fixed-width (+ 1 fixed-width width)))))
    (- (window-width) fixed-width)))

(defun slime-sprofile-browser-insert-columns (column-fn)
  "Call COLUMN-FN for each column in the sprofile browser layout.
Insert a space between every call to COLUMN-FN."
  (do ((layouts slime-sprofile-browser-column-layout (cdr layouts)))
      ((null layouts))
    (let ((layout (car layouts)))
      (let ((name (car layout))
            (position (cadr layout))
            (width (caddr layout)))
        (funcall column-fn
                 (slime-sprofile-browser-lookup-column name)
                 position
                 (if (eq width 'variable)
                     slime-sprofile-browser-fill-width
                     width))))
    (if (consp (cdr layouts))
        (insert " ")))
  (insert ?\n))

(defun slime-sprofile-browser-insert-heading-row ()
  (slime-sprofile-browser-insert-columns
   (lambda (column title-position width)
     (insert (slime-sprofile-browser-justify
              (slime-sprofile-browser-column-header column)
              title-position
              width)))))

(defun slime-sprofile-browser-insert-node-row (node)
  (slime-propertize-region
      `(,@(slime-sprofile-browser-node-properties
           (slime-sprofile-node-id node))
        ,@(slime-sprofile-browser-collapsed-properties))
    (slime-sprofile-browser-insert-columns
     (lexical-let ((node node))
       (lambda (column title-position width)
         (ignore title-position)
         (funcall (slime-sprofile-browser-column-formatter column)
                  (nth (slime-sprofile-browser-column-index column)
                       node)
                  width))))))

(defun slime-sprofile-browser-insert-integer (integer width)
  (insert (slime-sprofile-browser-format "%%-%dd" width integer)))

(defun slime-sprofile-browser-insert-name (name width)
  (slime-insert-propertized
   (slime-sprofile-browser-name-properties)
   (slime-sprofile-browser-format "%%-%ds" width name)))

(defun slime-sprofile-browser-insert-percent (value width)
  (insert (slime-sprofile-browser-format "%%%d.2f" width value)))

(defun slime-sprofile-browser-justify (string position max-width)
  (let ((width (length string)))
    (cond ((> width max-width)
           (substring string 0 max-width))
          ((eq position 'left)
           (concat string (make-string (- max-width width) ?\s)))
          ((eq position 'right)
           (concat (make-string (- max-width width) ?\s) string))
          ((eq position 'centre)
           (let ((padding
                  (make-string (floor (/ (- max-width width) 2))
                               ?\s)))
             (concat padding string padding
                     (if (oddp width) " " "")))))))

(defun slime-sprofile-browser-truncate (string width)
  (if (> (length string) width)
      (substring string 0 width)
      string))

(defun slime-sprofile-browser-format (meta-format width &rest args)
  (slime-sprofile-browser-truncate
   (apply 'format (format meta-format width) args)
   width))

;;;;; Browser Commands

(defun slime-sprofile-browser-redraw ()
  "Redraw the current sprofile browser buffer."
  (interactive)
  (slime-sprofile-browser-check)
  (let ((graph slime-sprofile-browser-graph))
    (setq slime-sprofile-browser-graph
          (slime-sprofile-browser-sort-graph graph)))
  (let ((inhibit-read-only t))
    (erase-buffer)
    (setq slime-sprofile-browser-fill-width
          (slime-sprofile-browser-calculate-fill))
    (slime-sprofile-browser-insert-heading-row)
    (dolist (node slime-sprofile-browser-graph)
      (slime-sprofile-browser-insert-node-row node)))
  (goto-line 2))

(defun slime-sprofile-browser-disassemble ()
  "Disassemble the code block at point."
  (interactive)
  (let ((id (slime-sprofile-browser-id-at-point)))
    (if id
        (slime-eval-describe `(swank:swank-sprofile-disassemble ,id))
        (slime-sprofile-browser-complain
         "Not on a code block to disassemble"))))

(defun slime-sprofile-browser-follow ()
  "Follow the code block name at point to its top-level entry."
  (interactive)
  (let ((id (slime-sprofile-browser-id-at-point)))
    (if id
        (let ((position
               (slime-sprofile-browser-find-position-by-id id)))
          (if position
              (goto-char position)
              (slime-sprofile-browser-complain
               "Unable to find top-level entry: %S"
               (slime-sprofile-browser-name-at-point))))
        (slime-sprofile-browser-complain
         "Not on a code block reference to follow"))))

(defun slime-sprofile-browser-sort (column)
  "Sort the sprofile browser listing."
  (interactive
   (list
    (slime-sprofile-browser-read-column-name "Order by column: ")))
  (slime-sprofile-browser-check)
  (setq slime-sprofile-browser-order-column column)
  (slime-sprofile-browser-redraw))

(defun slime-sprofile-browser-view-source ()
  "View the source code of the code block at point."
  (let ((id (slime-sprofile-browser-id-at-point)))
    (if (not id)
        (slime-sprofile-browser-complain
         "Not on a code block whose source to view")
        (slime-eval-async `(swank:swank-sprofile-source-location ,name)
          (lambda (source-location)
            (destructure-case source-location
              ((:error message)
               (slime-sprofile-browser-complain "%s" message))
              (t
               (slime-show-source-location source-location))))))))

;;;;; Expanding and Collapsing

(defun slime-sprofile-browser-toggle ()
  "Expand or collapse the entry for the code block at point."
  (interactive)
  (let ((id (slime-sprofile-browser-id-at-point)))
    (if id
        (if (slime-sprofile-browser-collapsed-at-point-p)
            (slime-sprofile-browser-expand id)
            (slime-sprofile-browser-collapse))
        (slime-sprofile-browser-complain "No code block at point"))))

(defun slime-sprofile-browser-collapse ()
  (let ((inhibit-read-only t)
        (depth (slime-sprofile-browser-depth-at-point)))
    (save-excursion
      (goto-char (1+ (point-at-eol)))
      (while (not
              (or (eobp)
                  (>= depth (slime-sprofile-browser-depth-at-point))))
        (delete-region (point-at-bol)
                       (min (1+ (point-at-eol))
                            (point-max)))))
    (slime-sprofile-browser-collapsed)))

(defun slime-sprofile-browser-expand (id)
  (slime-eval-async `(swank:swank-sprofile-expand-node ,id)
    (lexical-let ((id id)
                  (buffer (current-buffer))
                  (point (point))
                  (depth
                   (1+ (or (slime-sprofile-browser-depth-at-point)
                           0))))
      (lambda (edges)
        (with-current-buffer buffer
          (let ((inhibit-read-only t))
            (save-excursion
              (goto-char point)
              (slime-sprofile-browser-expanded)
                ;; The buffer ends with a newline, so this is safe.
              (goto-char (1+ (point-at-eol)))
              (slime-sprofile-browser-insert-expansion id
                                                       edges
                                                       depth))))))))

(defun slime-sprofile-browser-insert-expansion (id edges depth)
  (destructuring-bind (&key callers calls) edges
    (slime-propertize-region
        (slime-sprofile-browser-depth-properties id)
      (slime-sprofile-browser-insert-edges callers "Callers" depth)
      (slime-sprofile-browser-insert-edges calls "Callees" depth))))

(defun slime-sprofile-browser-insert-edges (edges title depth)
  (when (not (null edges))
    (insert (slime-sprofile-browser-expansion-title-padding depth))
    (insert title)
    (insert ?\n)
    (dolist (edge edges)
      (slime-sprofile-browser-insert-edge edge depth))))

;++ Fix!

(defun slime-sprofile-browser-insert-edge (edge depth)
  (destructuring-bind (id name cumulative) edge
    (slime-propertize-region
        `(,@(slime-sprofile-browser-node-properties id)
          ,@(slime-sprofile-browser-collapsed-properties))
      (insert "LOSE\n"))))

(defun slime-sprofile-browser-expansion-title-padding (depth)
  (make-string (* 2 depth) ?\s))

(defun slime-sprofile-browser-expansion-padding (depth)
  (make-string (+ 1 (* 2 depth)) ?\s))

(defun slime-sprofile-browser-node-properties (id)
  `(slime-sprofile-browser-node-id      ,id))

(defun slime-sprofile-browser-collapsed-properties ()
  '(slime-sprofile-browser-collapsed-p  t))

(defun slime-sprofile-browser-depth-properties (depth)
  `(slime-sprofile-browser-depth        ,depth))

(defun slime-sprofile-browser-expanded ()
  (add-text-properties (point-at-bol)
                       (point-at-eol)
                       '(slime-sprofile-browser-collapsed-p nil)))

(defun slime-sprofile-browser-collapsed ()
  (add-text-properties (point-at-bol)
                       (point-at-eol)
                       '(slime-sprofile-browser-collapsed-p t)))

(defun slime-sprofile-browser-collapsed-at-point-p ()
  (get-text-property (point) 'slime-sprofile-browser-collapsed-p))

(defun slime-sprofile-browser-id-at-point ()
  (get-text-property (point) 'slime-sprofile-browser-node-id))

(defun slime-sprofile-browser-find-position-by-id (id)
  (text-property-any (point-min)
                     (point-max)
                     'slime-sprofile-browser-node-id
                     id))

(defun slime-sprofile-browser-depth-at-point ()
  (or (get-text-property (point) 'slime-sprofile-browser-depth) 0))

(defun slime-sprofile-browser-name-properties ()
  '(face sldb-restart-number-face))

(defun slime-sprofile-browser-read-column-name (prompt)
  (intern (completing-read
           prompt
           (mapcar (lambda (column)
                     (symbol-name
                      (slime-sprofile-browser-column-name column)))
                   slime-sprofile-browser-column-descriptors)
           nil                          ;No predicate
           t)))                         ;Require a match

(defun slime-sprofile-browser-complain (format &rest args)
  (apply 'message format args)
  (ding))

(provide 'slime-sprof)

;;; Local Variables:
;;; Eval: (put 'slime-eval-async 'lisp-indent-hook 1)
;;; End:
